import asyncio
import logging
import subprocess
import threading
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import BinaryIO

from pygls.server import aio_readline

logger = logging.getLogger("lsp_devtools.agent")


def forward_message(source: str, dest: BinaryIO, message: bytes):
    """Forward the given message to the destination channel"""
    dest.write(message)
    dest.flush()

    # Log the full message
    logger.info(
        "%s",
        message.decode("utf8"),
        extra={"source": source},
    )


async def check_server_process(
    server_process: subprocess.Popen, stop_event: threading.Event
):
    """Ensure that the server process is still alive."""

    while not stop_event.is_set():
        retcode = server_process.poll()
        print(".")
        if retcode is not None:

            # Cancel any pending tasks.
            for task in asyncio.all_tasks():
                task.cancel(f"Server process exited with code: {retcode}")

            # Signal everything to stop.
            stop_event.set()

        await asyncio.sleep(0.1)


class Agent:
    """The Agent sits between a language server and its client, listening to messages
    enabling them to be recorded."""

    def __init__(self, server: subprocess.Popen, stdin: BinaryIO, stdout: BinaryIO):
        self.stdin = stdin
        self.stdout = stdout
        self.server_process = server
        self.stop_event = threading.Event()
        self.thread_pool_executor = ThreadPoolExecutor(
            max_workers=4,
            thread_name_prefix="LSP Traffic Worker ",
        )

    async def start(self):
        event_loop = asyncio.get_running_loop()

        # Connect stdin to the subprocess' stdin
        client_to_server = aio_readline(
            loop=event_loop,
            executor=self.thread_pool_executor,
            stop_event=self.stop_event,
            rfile=self.stdin,
            proxy=partial(forward_message, "client", self.server_process.stdin),
        )

        # Connect the subprocess' stdout to stdout
        server_to_client = aio_readline(
            loop=event_loop,
            executor=self.thread_pool_executor,
            stop_event=self.stop_event,
            rfile=self.server_process.stdout,
            proxy=partial(forward_message, "server", self.stdout),
        )

        # Run both connections concurrently.
        return await asyncio.gather(
            client_to_server,
            server_to_client,
            check_server_process(self.server_process, self.stop_event),
        )

    def stop(self):
        self.stop_event.set()
        self.thread_pool_executor.shutdown(wait=False, cancel_futures=True)

        try:
            self.server_process.terminate()
            ret = self.server_process.wait(timeout=1)
            print(f"Server process exited with code: {ret}")
        except TimeoutError:
            self.server_process.kill()

        # Need to close these to prevent open file warnings
        if self.server_process.stdin is not None:
            self.server_process.stdin.close()

        if self.server_process.stdout is not None:
            self.server_process.stdout.close()



import asyncio
import logging
import typing
from functools import partial

from textual.message import Message

from lsp_devtools.agent import MESSAGE_TEXT_NOTIFICATION
from lsp_devtools.agent import AgentClient
from lsp_devtools.agent import MessageText
from lsp_devtools.agent import parse_rpc_message
from lsp_devtools.record import logger
from lsp_devtools.record import setup_sqlite_output

if typing.TYPE_CHECKING:
    from . import LSPInspector


class Ping(Message):
    """Sent when the UI needs a refresh."""


class TUIAgentClient(AgentClient):
    def __init__(self, app: "LSPInspector"):
        self.app = app
        super().__init__()


def log_message(ls: TUIAgentClient, source: str, message: dict):
    logger.info("%s", message, extra={"source": source})
    app = ls.app

    # The event loop only becomes available once the on_ready event has fired
    # and the UI has bootstrapped itself. So it's possible we recevie
    # messages before the app is ready to accept them.
    if app.loop is None:
        return

    asyncio.run_coroutine_threadsafe(app.post_message(Ping(app)), app.loop)


def recv_message(ls: TUIAgentClient, message: MessageText):
    logfn = partial(log_message, ls, message.source)
    parse_rpc_message(ls, message, logfn)


def connect_to_agent(args, app: "LSPInspector"):
    client = TUIAgentClient(app)
    client.feature(MESSAGE_TEXT_NOTIFICATION)(recv_message)

    logger.setLevel(logging.INFO)
    setup_sqlite_output(args)

    return client



from prometheus_client import Counter  # type: ignore
from prometheus_client import Histogram  # type: ignore
from prometheus_client import start_http_server  # type: ignore

from lsp_devtools.handlers import LspHandler
from lsp_devtools.handlers import LspMessage

REQUESTS_COUNTER = Counter(
    "lsp_request_count", "Number of requests sent.", ["session", "source", "method"]
)
REQUESTS_DURATION = Histogram(
    "lsp_request_duration",
    "Time it takes to process a request",
    ["session", "source", "method"],
)
NOTIFICATIONS_COUNTER = Counter(
    "lsp_notification_count",
    "Number of notification sent.",
    ["session", "source", "method"],
)


class PrometheusHandler(LspHandler):
    """Logging handler that exports metrics to prometheus."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        start_http_server(8010)

        self.pending_requests = {}

    def handle_message(self, message: LspMessage):

        if message.is_request:
            REQUESTS_COUNTER.labels(
                message.session, message.source, message.method
            ).inc()
            self.pending_requests[message.id] = message

        if message.is_response:
            request = self.pending_requests.get(message.id, None)
            if request:
                REQUESTS_DURATION.labels(
                    request.session, request.source, request.method
                ).observe(message.timestamp - request.timestamp)

        if message.is_notification:
            NOTIFICATIONS_COUNTER.labels(
                message.session, message.source, message.method
            ).inc()



from pygls.protocol import default_converter
from pygls.server import Server

from lsp_devtools.agent.protocol import AgentProtocol


class AgentServer(Server):
    """A pygls server that wraps an agent allowing other processes to interact with it
    via websockets."""

    lsp: AgentProtocol

    def __init__(self, *args, **kwargs):

        if "protocol_cls" not in kwargs:
            kwargs["protocol_cls"] = AgentProtocol

        if "converter_factory" not in kwargs:
            kwargs["converter_factory"] = default_converter

        super().__init__(*args, **kwargs)



import argparse
import importlib
import logging
import sys
import traceback

from lsp_devtools import __version__

logger = logging.getLogger(__name__)


BUILTIN_COMMANDS = [
    "lsp_devtools.agent",
    "lsp_devtools.cmds.capabilities",  # TODO: Remove in favour of record + cli args
    "lsp_devtools.record",
    "lsp_devtools.tui",
]


def load_command(commands: argparse._SubParsersAction, name: str):
    try:
        mod = importlib.import_module(name)
    except Exception:
        logger.warning("Unable to load command '%s'\n%s", name, traceback.format_exc())
        return

    if not hasattr(mod, "cli"):
        logger.warning("Unable to load command '%s': missing 'cli' definition", name)
        return

    try:
        mod.cli(commands)
    except Exception:
        logger.warning("Unable to load command '%s'\n%s", name, traceback.format_exc())
        return


def main():

    cli = argparse.ArgumentParser(
        prog="lsp-devtools", description="Development tooling for language servers"
    )
    cli.add_argument("--version", action="version", version=f"%(prog)s v{__version__}")
    commands = cli.add_subparsers(title="commands")

    for mod in BUILTIN_COMMANDS:
        load_command(commands, mod)

    try:
        idx = sys.argv.index("--")
        args, extra = sys.argv[1:idx], sys.argv[idx + 1 :]
    except ValueError:
        args, extra = sys.argv[1:], None

    parsed_args = cli.parse_args(args)

    if hasattr(parsed_args, "run"):
        return parsed_args.run(parsed_args, extra)

    cli.print_help()
    return 0



import argparse
import json

from lsprotocol.types import INITIALIZE
from lsprotocol.types import InitializeParams
from pygls.server import LanguageServer


def capabilities(args, extra):

    server = LanguageServer(name="capabilities-dumper", version="v1.0")

    @server.feature(INITIALIZE)
    def on_initialize(ls: LanguageServer, params: InitializeParams):
        client_info = params.client_info
        if client_info:
            client_name = client_info.name.lower().replace(" ", "_")
            client_version = client_info.version or "unknown"
        else:
            client_name = "unknown"
            client_version = "unknown"

        filename = f"{client_name}_v{client_version}.json"
        with open(filename, "w") as f:
            obj = params.capabilities
            json.dump(ls.lsp._converter.unstructure(obj), f, indent=2)

    server.start_io()


def cli(commands: argparse._SubParsersAction):
    cmd = commands.add_parser(
        "capabilities",
        help="dummy lsp server for recording a client's capabilities.",
    )
    cmd.set_defaults(run=capabilities)



import logging
from typing import Dict
from typing import Literal
from typing import Set
from typing import Union

import attrs

from .formatters import FormatString

logger = logging.getLogger(__name__)

MessageSource = Literal["client", "server", "both"]
MessageType = Literal["request", "response", "result", "error", "notification"]


@attrs.define
class LSPFilter(logging.Filter):
    """Logging filter for LSP messages."""

    message_source: MessageSource = attrs.field(default="both")
    """Only include messages from the given source."""

    include_message_types: Set[MessageType] = attrs.field(factory=set, converter=set)
    """Only include the given message types."""

    exclude_message_types: Set[MessageType] = attrs.field(factory=set, converter=set)
    """Exclude the given message types."""

    include_methods: Set[str] = attrs.field(factory=set, converter=set)
    """Only include messages associated with the given method."""

    exclude_methods: Set[str] = attrs.field(factory=set, converter=set)
    """Exclude messages associated with the given method."""

    formatter: FormatString = attrs.field(
        default="",
        converter=FormatString,
    )  # type: ignore
    """Format messages according to the given string"""

    _response_method_map: Dict[Union[int, str], str] = attrs.field(factory=dict)
    """Used to determine the method for response messages"""

    def filter(self, record: logging.LogRecord) -> bool:
        message = record.args
        if not isinstance(message, dict):
            return False

        source = record.__dict__["source"]
        message_type = get_message_type(message)
        message_method = self._get_message_method(message_type, message)

        if self.message_source != "both" and source != self.message_source:
            return False

        if self.include_message_types and not message_matches_type(
            message_type, self.include_message_types
        ):
            return False

        if self.exclude_message_types and message_matches_type(
            message_type, self.exclude_message_types
        ):
            return False

        if self.include_methods and message_method not in self.include_methods:
            return False

        if self.exclude_methods and message_method in self.exclude_methods:
            return False

        if self.formatter.pattern:
            try:
                record.msg = self.formatter.format(message)
            except Exception:
                logger.debug(
                    "Skipping message that failed to format: %s", message, exc_info=True
                )
                return False

        return True

    def _get_message_method(self, message_type: str, message: dict) -> str:

        if message_type == "request":
            method = message["method"]
            self._response_method_map[message["id"]] = method

            return method

        if message_type == "notification":
            return message["method"]

        return self._response_method_map[message["id"]]


def message_matches_type(message_type: str, types: Set[MessageType]) -> bool:
    """Determine if the type of message is included in the given set of types"""

    if message_type == "result":
        return len({"result", "response"} & types) > 0

    if message_type == "error":
        return len({"error", "response"} & types) > 0

    return message_type in types


def get_message_type(message: dict) -> str:
    if "id" in message:
        if "error" in message:
            return "error"
        elif "method" in message:
            return "request"
        else:
            return "result"
    else:
        return "notification"



import re
from functools import cache
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import lsprotocol.types


def format_position(position: dict) -> str:
    return f"{position['line']}:{position['character']}"


def format_range(range_: dict) -> str:
    return f"{format_position(range_['start'])}-{format_position(range_['end'])}"


FORMATTERS = {
    "position": format_position,
    "range": format_range,
}


def get_formatter(fmt: str) -> Callable[[Any], str]:
    """Return the formatter with the given name."""

    formatter = FORMATTERS.get(fmt.lower(), None)
    if formatter is not None:
        return formatter

    # Is the formatter is an enum?
    enum = getattr(lsprotocol.types, fmt, None)
    if enum is not None:

        def fn(v: int) -> str:
            return enum(v).name  # type: ignore

        return fn

    raise TypeError(f"Unknown format: '{fmt}'")


class Value:
    """Represents a value to be inserted into a string."""

    LIST_FIELD = re.compile(r"(.*)\[(.*)?\]")

    def __init__(self, accessor: str, formatter: Callable[[Any], str]):
        self.accessor = accessor
        self.formatter = formatter

    def __repr__(self):
        return f'Value(accessor="{self.accessor}", formatter={self.formatter})'

    def format(self, message: dict, accessor: Optional[str] = None) -> str:
        """Convert a message to a string according to the current accessor and
        formatter."""

        idx = 0
        obj = message
        accessor = self.accessor if accessor is None else accessor

        for field in accessor.split("."):

            if not field:
                continue

            # TODO: Once we have a better idea on what we want the list syntax to be,
            # most of this should be move into the format string parsing code, so that
            # it only runs once.
            idx += len(field)
            match = self.LIST_FIELD.fullmatch(field)
            if match:
                obj = obj[match.group(1)]
                remainder = accessor[idx + 1 :]
                sep, index = get_separator_index(match.group(2))

                if isinstance(index, int):
                    obj = obj[index]
                    continue
                elif isinstance(index, slice):
                    obj = obj[index]

                return sep.join([self.format(o, remainder) for o in obj])

            obj = obj[field]

        return self.formatter(obj)


@cache
def get_separator_index(separator: str) -> Tuple[str, Union[int, slice, None]]:
    if not separator:
        return "\n", None

    if "#" in separator:
        idx, sep = separator.split("#")
        return get_separator(sep), get_index(idx)

    # Try parsing as an index
    index = get_index(separator)
    if index is not None:
        return "\n", index

    # It must be just a separator then.
    return get_separator(separator), None


def get_separator(sep: str) -> str:
    if not sep:
        return "\n"

    return sep.replace("\\n", "\n").replace("\\t", "\t")


def get_index(idx: str) -> Union[int, slice, None]:

    try:
        return int(idx)
    except ValueError:
        pass

    if ":" in idx:
        parts = idx.split(":")

        try:
            return slice(*[int(p) for p in parts])
        except ValueError:
            pass

    return None


class FormatString:
    """Implements the format string syntax.

    [{.params.type:MessageType}]: {.params.message} and something else
    """

    VARIABLE = re.compile(r"{\.([^}]+)}")

    def __init__(self, pattern: str):
        self.pattern = pattern
        self._parse()

    def _parse(self):
        idx = 0
        parts: List[Union[str, Value]] = []

        for match in self.VARIABLE.finditer(self.pattern):
            start, end = match.span()
            parts.append(self.pattern[idx:start])

            variable = match.group(1)
            if "|" in variable:
                accessor, fmt = variable.split("|")
                formatter = get_formatter(fmt)
            else:
                accessor = variable
                formatter = str

            parts.append(Value(accessor=accessor, formatter=formatter))
            idx = end

        parts.append(self.pattern[idx:])
        self.parts = parts

    def format(self, message: dict) -> str:
        return "".join([p.format(message) for p in self.parts])



import asyncio
import json
import re
from json.decoder import JSONDecodeError
from threading import Event
from typing import Any
from typing import Callable
from typing import Optional

import websockets
from pygls.protocol import default_converter
from pygls.server import Server
from websockets.client import WebSocketClientProtocol

from lsp_devtools.agent.protocol import AgentProtocol
from lsp_devtools.agent.protocol import MessageText


class WebSocketClientTransportAdapter:
    """Protocol adapter for the WebSocket client interface."""

    def __init__(self, ws: WebSocketClientProtocol, loop: asyncio.AbstractEventLoop):
        self._ws = ws
        self._loop = loop

    def close(self) -> None:
        """Stop the WebSocket server."""
        print("-- CLOSING --")
        self._loop.create_task(self._ws.close())

    def write(self, data: Any) -> None:
        """Create a task to write specified data into a WebSocket."""
        asyncio.ensure_future(self._ws.send(data))


MESSAGE_PATTERN = re.compile(
    r"^(?:[^\r\n]+\r\n)*"
    + r"Content-Length: (?P<length>\d+)\r\n"
    + r"(?:[^\r\n]+\r\n)*\r\n"
    + r"(?P<body>{.*)",
    re.DOTALL,
)


def parse_rpc_message(
    ls: "AgentClient", message: MessageText, callback: Callable[[dict], None]
):
    """Parse json-rpc messages coming from the agent.

    Originally adatped from the ``data_received`` method on pygls' ``JsonRPCProtocol``
    class.
    """
    data = message.text
    message_buf = ls._client_buf if message.source == "client" else ls._server_buf

    while len(data):

        # Append the incoming chunk to the message buffer
        message_buf.append(data)

        # Look for the body of the message
        msg = "".join(message_buf)
        found = MESSAGE_PATTERN.fullmatch(msg)

        body = found.group("body") if found else ""
        length = int(found.group("length")) if found else 1

        if len(body) < length:
            # Message is incomplete; bail until more data arrives
            return

        # Message is complete;
        # extract the body and any remaining data,
        # and reset the buffer for the next message
        body, data = body[:length], body[length:]
        message_buf.clear()

        callback(json.loads(body))


class AgentClient(Server):
    """Client for connecting to an LSPAgent instance."""

    lsp: AgentProtocol

    def __init__(self):
        super().__init__(
            protocol_cls=AgentProtocol, converter_factory=default_converter
        )
        self._client_buf = []
        self._server_buf = []
        self._stop_event: Event = Event()

    def _report_server_error(self, error, source):
        # Bail on error
        # TODO: Report the actual error somehow
        self._stop_event.set()

    def feature(self, feature_name: str, options: Optional[Any] = None):
        return self.lsp.fm.feature(feature_name, options)

    def start_ws_client(self, host: str, port: int):
        """Similar to ``start_ws``, but where we create a client connection rather than
        host a server."""

        self.lsp._send_only_body = True  # Don't send headers within the payload

        async def client_connection(host: str, port: int):
            """Create and run a client connection."""

            self._client = await websockets.connect(  # type: ignore
                f"ws://{host}:{port}"
            )
            self.lsp.transport = WebSocketClientTransportAdapter(
                self._client, self.loop
            )
            message = None

            try:
                while not self._stop_event.is_set():
                    try:
                        message = await asyncio.wait_for(
                            self._client.recv(), timeout=0.5
                        )
                        self.lsp._procedure_handler(
                            json.loads(
                                message, object_hook=self.lsp._deserialize_message
                            )
                        )
                    except JSONDecodeError:
                        print(message or "-- message not found --")
                        raise
                    except TimeoutError:
                        pass
                    except Exception:
                        raise

            finally:
                await self._client.close()

        try:
            asyncio.run(client_connection(host, port))
        except KeyboardInterrupt:
            pass
        finally:
            self.shutdown()



import argparse
import asyncio
import json
import pathlib
import threading
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import aiosqlite
import appdirs
from rich.highlighter import ReprHighlighter
from rich.text import Text
from textual import events
from textual.app import App
from textual.app import ComposeResult
from textual.containers import Container
from textual.events import Ready
from textual.widgets import DataTable
from textual.widgets import Footer
from textual.widgets import Header
from textual.widgets import Tree
from textual.widgets import TreeNode

from lsp_devtools.record import setup_filter_args

from .client import Ping
from .client import TUIAgentClient
from .client import connect_to_agent


class ObjectViewer(Tree):
    """Used to inspect the fields of an object."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.highlighter = ReprHighlighter()

    def set_object(self, name: str, obj: Any):
        self.clear()
        self.root.set_label(name)
        self.walk_object(name, self.root, obj)

    def walk_object(self, label: str, node: TreeNode, obj: Any):

        if isinstance(obj, dict):
            node.expand()
            for field, value in obj.items():
                child = node.add(field)
                self.walk_object(field, child, value)

        elif isinstance(obj, list):
            node.expand()
            for idx, value in enumerate(obj):
                child_label = self.highlighter(str(idx))
                child = node.add(child_label)
                self.walk_object(str(idx), child, value)

        else:
            node.allow_expand = False
            node.set_label(Text.assemble(label, " = ", self.highlighter(repr(obj))))


RPCData = Dict[int, Tuple[Optional[str], Optional[str], Optional[str]]]


class MessagesTable(DataTable):
    """Datatable used to display all messages between client and server"""

    def __init__(self, dbpath: pathlib.Path, viewer: ObjectViewer):
        super().__init__()

        self.dbpath = dbpath
        self.dbquery = "SELECT rowid, * FROM protocol WHERE rowid > ?"
        self.rpcdata: RPCData = {}
        self.max_row = -1

        self.viewer = viewer

        self.add_column("")
        self.add_column("Time")
        self.add_column("Source")
        self.add_column("ID")
        self.add_column("Method")

    def on_key(self, event: events.Key):

        if event.key != "enter":
            return

        rowid = int(self.data[self.cursor_row][0])
        params, result, error = self.rpcdata[rowid]

        if params:
            name = "params"
            message = json.loads(params)

        elif result:
            name = "result"
            message = json.loads(result)

        elif error:
            name = "error"
            message = json.loads(error)

        else:
            name = "data"
            message = {}

        self.viewer.set_object(name, message)

    async def update(self):
        """Trigger a re-run of the query to pull in new data."""

        async with aiosqlite.connect(self.dbpath) as conn:
            async with conn.execute(self.dbquery, (self.max_row,)) as cursor:
                async for row in cursor:
                    rowid = row[0]
                    timestamp = row[2]
                    source = row[3]
                    id_ = row[4]
                    method = row[5]
                    params = row[6]
                    result = row[7]
                    error = row[8]

                    self.rpcdata[rowid] = (params, result, error)

                    # Surely there's a more direct way to do this?
                    dt = datetime.fromtimestamp(timestamp)
                    time = dt.isoformat(timespec="milliseconds")
                    time = time[time.find("T") + 1 :]

                    self.add_row(str(rowid), time, source, id_, method)
                    self.max_row = rowid


class Sidebar(Container):
    pass


class LSPInspector(App):

    CSS_PATH = pathlib.Path(__file__).parent / "app.css"
    BINDINGS = [("ctrl+b", "toggle_sidebar", "Sidebar"), ("q", "quit", "Quit")]

    def __init__(self, dbpath: pathlib.Path, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.dbpath = dbpath
        """Where the data for the app is being held"""

        self.client: Optional[TUIAgentClient] = None
        """Client used to interact with the LSPAgent hosting the server we're
        inspecting."""

        self.loop: Optional[asyncio.AbstractEventLoop] = None
        """Accessed by the AgentClient to push messages into the UI"""

    def compose(self) -> ComposeResult:
        yield Header()

        viewer = ObjectViewer("")
        messages = MessagesTable(self.dbpath, viewer)

        yield Container(messages, Sidebar(viewer))
        yield Footer()

    def action_toggle_sidebar(self) -> None:

        sidebar = self.query_one(Sidebar)
        self.set_focus(None)

        if sidebar.has_class("-hidden"):
            sidebar.remove_class("-hidden")
        else:
            if sidebar.query("*:focus"):
                self.screen.set_focus(None)
            sidebar.add_class("-hidden")

    async def on_ready(self, event: Ready):
        self.loop = asyncio.get_running_loop()
        await self.update_table()

    async def on_ping(self, message: Ping):
        """Fired when the agent client receives new messages"""
        await self.update_table()

    async def update_table(self):
        table = self.query_one(MessagesTable)
        await table.update()

    async def action_quit(self):
        if self.client:
            self.client._stop_event.set()
        await super().action_quit()


def tui(args, extra: List[str]):

    dbpath = args.to_sqlite
    if not dbpath.parent.exists():
        dbpath.parent.mkdir(parents=True)

    app = LSPInspector(dbpath)
    client = connect_to_agent(args, app)
    app.client = client

    agent_thread = threading.Thread(
        name="AgentClient", target=client.start_ws_client, args=(args.host, args.port)
    )
    agent_thread.start()

    app.run()
    agent_thread.join()


def cli(commands: argparse._SubParsersAction):
    cmd: argparse.ArgumentParser = commands.add_parser(
        "tui",
        help="launch TUI",
        description="""\
This command opens a text user interface that can be used to inspect and
manipulate an LSP session interactively.
""",
    )

    default_db = pathlib.Path(
        appdirs.user_cache_dir(appname="lsp-devtools", appauthor="swyddfa"),
        "sessions.db",
    )
    cmd.add_argument(
        "--dbpath",
        type=pathlib.Path,
        metavar="DB",
        default=default_db,
        dest="to_sqlite",  # to be compatible with code borrowed from record command.
        help="the database path to use",
    )

    connect = cmd.add_argument_group(
        title="connection options",
        description="options that control the connection to the LSP Agent.",
    )
    connect.add_argument(
        "--host",
        type=str,
        default="localhost",
        help="the host that is hosting the agent.",
    )
    connect.add_argument(
        "-p", "--port", type=int, default=8765, help="the port to connect to."
    )

    setup_filter_args(cmd)
    cmd.set_defaults(run=tui)



import argparse
import logging
import pathlib
from typing import List
from typing import Optional

from pygls.protocol import partial
from rich.console import ConsoleRenderable
from rich.logging import RichHandler
from rich.traceback import Traceback

from lsp_devtools.agent import MESSAGE_TEXT_NOTIFICATION
from lsp_devtools.agent import AgentClient
from lsp_devtools.agent import MessageText
from lsp_devtools.agent import parse_rpc_message
from lsp_devtools.handlers.sql import SqlHandler

from .filters import LSPFilter

logger = logging.getLogger(__name__)


class RichLSPHandler(RichHandler):
    def __init__(self, level: int, log_time_format="%X", **kwargs):
        super().__init__(
            level=level,
            show_path=False,
            log_time_format=log_time_format,
            omit_repeated_times=False,
            **kwargs,
        )

    def render(
        self,
        *,
        record: logging.LogRecord,
        traceback: Optional[Traceback],
        message_renderable: "ConsoleRenderable",
    ) -> "ConsoleRenderable":

        # Delegate most of the rendering to the base RichHandler class.
        res = super().render(
            record=record, traceback=traceback, message_renderable=message_renderable
        )

        # Abuse the log level column to display the source of the message,
        source = record.__dict__["source"]
        color = "red" if source == "client" else "blue"
        message_source = f"[bold][{color}]{source.upper()}[/{color}][/bold]"
        res.columns[1]._cells[0] = message_source  # type: ignore

        return res


def log_raw_message(ls: AgentClient, message: MessageText):
    """Push raw messages through the logging system."""
    logger.info(message.text, extra={"source": message.source})


def log_rpc_message(ls: AgentClient, message: MessageText):
    """Push parsed json-rpc messages through the logging system"""

    logfn = partial(logger.info, "%s", extra={"source": message.source})
    parse_rpc_message(ls, message, logfn)


def setup_stdout_output(args):
    """Log to stdout."""
    handler = RichLSPHandler(level=logging.INFO)
    handler.addFilter(
        LSPFilter(
            message_source=args.message_source,
            include_message_types=args.include_message_types,
            exclude_message_types=args.exclude_message_types,
            include_methods=args.include_methods,
            exclude_methods=args.exclude_methods,
            formatter=args.format_message,
        )
    )

    logger.addHandler(handler)


def setup_file_output(args):
    handler = logging.FileHandler(filename=str(args.to_file))
    handler.setLevel(logging.INFO)
    handler.addFilter(
        LSPFilter(
            message_source=args.message_source,
            include_message_types=args.include_message_types,
            exclude_message_types=args.exclude_message_types,
            include_methods=args.include_methods,
            exclude_methods=args.exclude_methods,
            formatter=args.format_message,
        )
    )

    logger.addHandler(handler)


def setup_sqlite_output(args):
    handler = SqlHandler(args.to_sqlite)
    handler.setLevel(logging.INFO)
    handler.addFilter(
        LSPFilter(
            message_source=args.message_source,
            include_message_types=args.include_message_types,
            exclude_message_types=args.exclude_message_types,
            include_methods=args.include_methods,
            exclude_methods=args.exclude_methods,
        )
    )

    logger.addHandler(handler)


def start_recording(args, extra: List[str]):

    client = AgentClient()
    log_func = log_raw_message if args.capture_raw_output else log_rpc_message
    logger.setLevel(logging.INFO)
    client.feature(MESSAGE_TEXT_NOTIFICATION)(log_func)

    if args.to_file:
        setup_file_output(args)

    elif args.to_sqlite:
        setup_sqlite_output(args)

    else:
        setup_stdout_output(args)

    try:
        client.start_ws_client(args.host, args.port)
    except Exception:
        # TODO: Error handling
        raise


def setup_filter_args(cmd: argparse.ArgumentParser):
    """Add arguments that can be used to filter messages."""

    filter = cmd.add_argument_group(
        title="filter options",
        description=(
            "select which messages to record, mutliple options will be ANDed together. "
            "Does not apply to raw message capture"
        ),
    )
    filter.add_argument(
        "--message-source",
        default="both",
        choices=["client", "server", "both"],
        help="only include messages from the given source",
    )
    filter.add_argument(
        "--include-message-type",
        action="append",
        default=[],
        dest="include_message_types",
        choices=["request", "response", "result", "error", "notification"],
        help="only include the given message type(s)",
    )
    filter.add_argument(
        "--exclude-message-type",
        action="append",
        dest="exclude_message_types",
        default=[],
        choices=["request", "response", "result", "error", "notification"],
        help="omit the given message type(s)",
    )
    filter.add_argument(
        "--include-method",
        action="append",
        dest="include_methods",
        default=[],
        metavar="METHOD",
        help="only include the given messages for the given method(s)",
    )
    filter.add_argument(
        "--exclude-method",
        action="append",
        dest="exclude_methods",
        default=[],
        metavar="METHOD",
        help="omit messages for the given method(s)",
    )


def cli(commands: argparse._SubParsersAction):
    cmd: argparse.ArgumentParser = commands.add_parser(
        "record",
        help="record an LSP session, requires the server be wrapped by an agent.",
        description="""\
This command connects to an LSP agent allowing for messages sent
between client and server to be logged.
""",
    )

    connect = cmd.add_argument_group(
        title="connection options", description="how to connect to the LSP agent"
    )
    connect.add_argument(
        "--host",
        type=str,
        default="localhost",
        help="the host that is hosting the agent.",
    )
    connect.add_argument(
        "-p", "--port", type=int, default=8765, help="the port to connect to."
    )

    capture = cmd.add_mutually_exclusive_group()
    capture.add_argument(
        "--capture-raw-output",
        action="store_true",
        help="capture the raw output from client and server.",
    )
    capture.add_argument(
        "--capture-rpc-output",
        default=True,
        action="store_true",
        help="capture the rpc messages sent between client and server.",
    )

    setup_filter_args(cmd)
    format = cmd.add_argument_group(
        title="formatting options",
        description=(
            "control how the recorded messages are formatted "
            "(does not apply to SQLite output or raw message capture)"
        ),
    )
    format.add_argument(
        "-f",
        "--format-message",
        default="",
        help=(
            "format messages according to given format string, "
            "see example commands above for syntax. "
            "Messages which fail to format will be excluded"
        ),
    )

    output = cmd.add_argument_group(
        title="output options",
        description="control where the captured messages are sent to",
    )
    output.add_argument(
        "--to-file",
        default=None,
        metavar="FILE",
        type=pathlib.Path,
        help="save messages to a file",
    )
    output.add_argument(
        "--to-sqlite",
        default=None,
        metavar="FILE",
        type=pathlib.Path,
        help="save messages to a SQLite DB",
    )

    cmd.set_defaults(run=start_recording)


def _enable_pygls_logging():
    pygls_log = logging.getLogger("pygls")
    pygls_log.setLevel(logging.DEBUG)
    pygls_log.addHandler(RichHandler(level=logging.DEBUG))



import logging
from typing import Any
from typing import Literal
from typing import Mapping
from typing import Optional
from uuid import uuid4

import attrs

MessageSource = Literal["client", "server"]


@attrs.define
class LspMessage:
    """A container that holds a message from the LSP protocol, with some additional
    metadata."""

    Source = MessageSource

    session: str
    """An id representing the session the message is a part of."""

    timestamp: float
    """When the message was sent."""

    source: MessageSource
    """Indicates if the message was sent by the client or the server."""

    id: Optional[str]
    """The ``id`` field, if it exists."""

    method: Optional[str]
    """The ``method`` field, if it exists."""

    params: Optional[Any]
    """The ``params`` field, if it exists."""

    result: Optional[Any]
    """The ``result`` field, if it exists."""

    error: Optional[Any]
    """The ``error`` field, if it exists."""

    @classmethod
    def from_rpc(
        cls, session: str, timestamp: float, source: str, message: Mapping[str, Any]
    ):
        """Create an instance from a JSON-RPC message."""
        return cls(
            session=session,
            timestamp=timestamp,
            source=source,  # type: ignore
            id=message.get("id", None),
            method=message.get("method", None),
            params=message.get("params", None),
            result=message.get("result", None),
            error=message.get("error", None),
        )

    @property
    def is_request(self) -> bool:
        return self.id is not None and self.params is not None

    @property
    def is_response(self) -> bool:
        result_or_error = self.result is not None or self.error is not None
        return self.id is not None and result_or_error

    @property
    def is_notification(self) -> bool:
        return self.id is None and self.params is not None


class LspHandler(logging.Handler):
    """Base class for lsp log handlers."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session_id = ""

    def handle_message(self, message: LspMessage):
        """Called each time a message is processed."""

    def emit(self, record: logging.LogRecord):

        if not isinstance(record.args, dict):
            return

        message = record.args
        timestamp = record.created
        source = record.__dict__["source"]

        if message.get("method", None) == "initialize":
            self.session_id = str(uuid4())

        self.handle_message(
            LspMessage.from_rpc(
                session=self.session_id,
                timestamp=timestamp,
                source=source,
                message=message,
            )
        )



from typing import Optional
from typing import Type

import attrs
from pygls.protocol import JsonRPCProtocol

MESSAGE_TEXT_NOTIFICATION = "message/text"


@attrs.define
class MessageText:
    """The contents of a ``message/text`` notification."""

    text: str
    """The captured text."""

    source: str
    """The source the text was captured from e.g. client."""


@attrs.define
class MessageTextNotification:
    """Notify the client of output that was captured by the agent."""

    jsonrpc: str
    method: str
    params: MessageText = attrs.field()


class AgentProtocol(JsonRPCProtocol):
    """The RPC protocol exposed by the agent."""

    def get_message_type(self, method: str) -> Optional[Type]:

        if method == MESSAGE_TEXT_NOTIFICATION:
            return MessageTextNotification

        return super().get_message_type(method)

    def message_text_notification(self, message: MessageText):
        """Send an ``message/text`` notification to the client."""
        self.notify(MESSAGE_TEXT_NOTIFICATION, message)



import argparse
import asyncio
import logging
import subprocess
import sys
import threading
from typing import List

from .agent import Agent
from .agent import logger
from .client import AgentClient
from .client import parse_rpc_message
from .protocol import MESSAGE_TEXT_NOTIFICATION
from .protocol import MessageText
from .server import AgentServer

__all__ = [
    "Agent",
    "AgentClient",
    "AgentServer",
    "logger",
    "MESSAGE_TEXT_NOTIFICATION",
    "MessageText",
    "parse_rpc_message",
]


class WSHandler(logging.Handler):
    """Logging handler that forwards captured LSP messages through to the web socket
    client."""

    def __init__(self, server: AgentServer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.server = server

    def emit(self, record: logging.LogRecord):
        message = MessageText(
            text=record.args[0],  # type: ignore
            source=record.__dict__["source"],
        )
        self.server.lsp.message_text_notification(message)


def run_agent(args, extra: List[str]):

    if extra is None:
        print("Missing server start command", file=sys.stderr)
        return 1

    process = subprocess.Popen(extra, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    agent = Agent(process, sys.stdin.buffer, sys.stdout.buffer)

    server = AgentServer()
    handler = WSHandler(server)
    handler.setLevel(logging.INFO)

    logger.setLevel(logging.INFO)
    logger.addHandler(handler)

    agent_thread = threading.Thread(
        target=asyncio.run,
        args=(agent.start(),),
    )
    agent_thread.start()

    try:
        server.start_ws(args.host, args.port)
    finally:
        agent.stop()


def cli(commands: argparse._SubParsersAction):
    cmd: argparse.ArgumentParser = commands.add_parser(
        "agent",
        help="instrument an LSP session",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""\
This command runs the given language server as a subprocess, wrapping it in a
websocket server allowing all traffic to be inspected by some client.

To wrap a server, supply its start command after all other agent options and
preceeded by a `--`, for example:

    lsp-devtools agent -p 1234 -- python -m esbonio

Wrapping a language server with this command is required to enable the
majority of the lsp-devtools suite of tools.

       ┌─ LSP Client ─┐     ┌─────── Agent ────────┐    ┌─ LSP Server ─┐
       │              │     │   ┌──────────────┐   │    │              │
       │        stdout│─────│───│              │───│────│stdin         │
       │              │     │   │ Agent Server │   │    │              │
       │         stdin│─────│───│              │───│────│stdout        │
       │              │     │   └──────────────┘   │    │              │
       │              │     │          │           │    │              │
       └──────────────┘     └──────────┴───────────┘    └──────────────┘
                                       │
                                       │ web socket
                                       │
                                ┌──────────────┐
                                │              │
                                │ Agent Client │
                                │              │
                                └──────────────┘

""",
    )

    cmd.add_argument(
        "--host",
        help="the host to run the websocket server on.",
        default="localhost",
    )
    cmd.add_argument(
        "-p",
        "--port",
        help="the port to run the websocket server on",
        default=8765,
    )

    cmd.set_defaults(run=run_agent)



import json
import pathlib
import sqlite3
import sys
from contextlib import closing

from lsp_devtools.handlers import LspHandler
from lsp_devtools.handlers import LspMessage

if sys.version_info.minor < 9:
    import importlib_resources as resources
else:
    import importlib.resources as resources  # type: ignore[no-redef]


class SqlHandler(LspHandler):
    """A logging handler that sends log records to a SQL database"""

    def __init__(self, dbpath: pathlib.Path, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.dbpath = dbpath

        resource = resources.files("lsp_devtools.handlers").joinpath("dbinit.sql")
        sql_script = resource.read_text(encoding="utf8")

        with closing(sqlite3.connect(self.dbpath)) as conn:
            conn.executescript(sql_script)

    def handle_message(self, message: LspMessage):

        with closing(sqlite3.connect(self.dbpath)) as conn:
            cursor = conn.cursor()
            cursor.execute(
                "INSERT INTO protocol VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    message.session,
                    message.timestamp,
                    message.source,
                    message.id,
                    message.method,
                    json.dumps(message.params) if message.params else None,
                    json.dumps(message.result) if message.result else None,
                    json.dumps(message.error) if message.error else None,
                ),
            )

            conn.commit()
